@@ -35,7 +35,6 @@
 #include <near/log.h>
 #include <near/types.h>
 #include <near/adapter.h>
-#include <near/target.h>
 #include <near/tag.h>
 #include <near/ndef.h>
 #include <near/tlv.h>
@@ -218,13 +217,13 @@ static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
 	memcpy(nfc_data + current_length, resp + OFS_READ_DATA, length_read);
 
 	if (current_length + length_read >= data_length) {
+		GList *records;
+
 		tag->current_block = 0;
 
 		DBG("Done reading %d bytes at %p", data_length, nfc_data);
-		near_ndef_parse(tag->tag, nfc_data, data_length);
-
-		if (tag->cb)
-			tag->cb(adapter_idx, target_idx, 0);
+		records = near_ndef_parse(nfc_data, data_length);
+		near_tag_add_records(tag->tag, records, tag->cb, 0);
 
 		g_free(tag);
 
@@ -305,9 +304,13 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	ndef_data_length *= 0x100;
 	ndef_data_length += resp[OFS_READ_DATA + 13];
 
-	/* Associate the DATA length to the tag */
-	tag = near_target_add_tag(cookie->adapter_idx, cookie->target_idx,
+	/* Add data to the tag */
+	err = near_tag_add_data(cookie->adapter_idx, cookie->target_idx,
 					NULL, ndef_data_length);
+	if (err < 0)
+		goto out;
+
+	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
 	if (tag == NULL) {
 		err = -ENOMEM;
 		goto out;
@@ -334,7 +337,6 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	t3_tag->cb = cookie->cb;
 	t3_tag->tag = tag;
 
-	near_tag_set_uid(tag, cookie->IDm , LEN_ID);
 	err = nfctype3_data_read(t3_tag);
 
 out:
@@ -388,7 +390,7 @@ out:
 	return err;
 }
 
-static int nfctype3_read_tag(uint32_t adapter_idx,
+static int nfctype3_read(uint32_t adapter_idx,
 				uint32_t target_idx, near_tag_io_cb cb)
 {
 	struct type3_cmd cmd;
@@ -599,7 +601,7 @@ out:
 	return err;
 }
 
-static int nfctype3_write_tag(uint32_t adapter_idx, uint32_t target_idx,
+static int nfctype3_write(uint32_t adapter_idx, uint32_t target_idx,
 				struct near_ndef_message *ndef,
 				near_tag_io_cb cb)
 {
@@ -610,7 +612,7 @@ static int nfctype3_write_tag(uint32_t adapter_idx, uint32_t target_idx,
 	if (ndef == NULL || cb == NULL)
 		return -EINVAL;
 
-	tag = near_target_get_tag(adapter_idx, target_idx);
+	tag = near_tag_get_tag(adapter_idx, target_idx);
 	if (tag == NULL)
 		return -EINVAL;
 
@@ -673,6 +675,8 @@ static int nfctype3_check_presence(uint32_t adapter_idx,
 	if (err < 0)
 		goto out;
 
+	return 0;
+
 out:
 	t3_cookie_release(cookie);
 
@@ -682,8 +686,8 @@ out:
 static struct near_tag_driver type1_driver = {
 	.type           = NFC_PROTO_FELICA,
 	.priority       = NEAR_TAG_PRIORITY_DEFAULT,
-	.read_tag       = nfctype3_read_tag,
-	.add_ndef       = nfctype3_write_tag,
+	.read           = nfctype3_read,
+	.write          = nfctype3_write,
 	.check_presence = nfctype3_check_presence,
 };
 
